import itertools
from .util import comp, GlueMergeSpec
import re
from .tiles import TileList
from .sensitivitynew import _fakesingles

_OE = [2, 3, 0, 1]
_ENDER = [3, 2, 1, 0]


def _generic_branch(direction, tiles, tile, n, f=False, gms=GlueMergeSpec([])):
    # Generic branch finder.  MUST USE ALL SINGLES (EG, FAKESINGLE)
    branches = []
    if n == 0:
        return [([tile], tile['ends'][_ENDER[direction]])]
    for tn in tiles:
        if gms.eq(tile['ends'][direction], comp(tn['ends'][_OE[direction]])):
            branches += [
                ([tile] + x, y)
                for x, y in _generic_branch(direction, tiles, tn, n - 1)
            ]
    return branches


def _latticedefect_tile(tiles, tile, direction='e', n=2,
                        gms=GlueMergeSpec([])):
    """With n tiles in each branch, can tile t form a lattice defect?"""
    d1, d2 = {'e': (1, 2), 'w': (0, 3), 'n': (0, 1), 's': (2, 3)}[direction]
    b1 = _generic_branch(d1, tiles, tile, n, gms=gms)
    b2 = _generic_branch(d2, tiles, tile, n, gms=gms)
    neighborhoods = itertools.product(b1, b2)
    res = []
    for n in neighborhoods:
        res += [(n, tile) for tile in tiles
                if (gms.eq(n[0][1], comp(tile['ends'][_OE[_ENDER[d1]]]))
                    and gms.eq(n[1][1], comp(tile['ends'][_OE[_ENDER[d2]]])))]
    return res


def _ppld(res):
    """Pretty-print lattice defects, to some extent.  I is the initial
    tile, A/B are the two branches, and F is the tile that
    attaches to those branches.
    """
    return [
        "I:" + n[0][0][0]['name'] + " " + "A:[" + ",".join(
            t['name'] for t in n[0][0][1:]) + "] " + "B:[" + ",".join(
                t['name'] for t in n[1][0][1:]) + "] " + "F:" + tt['name']
        for n, tt in res
    ]


def latticedefects(ts,
                   direction='e',
                   depth=2,
                   pp=True,
                   rotate=False,
                   gms=GlueMergeSpec([])):
    if depth < 2:
        raise ValueError(
            "Depth cannot be less than 2, received {}.".format(depth))
    tiles = _fakesingles(ts.tiles)
    rtiles = _fakesingles(
        TileList([x for x in ts.tiles if 'fake' not in x.keys()]) + sum([
            x.rotations for x in ts.tiles if 'fake' not in x.keys()
        ], TileList()))
    if rotate:
        tll = rtiles
    else:
        tll = tiles
    alldefects = sum((_latticedefect_tile(
        tll, tile, direction=direction, n=depth, gms=gms) for tile in tll), [])
    if pp:
        return _ppld(alldefects)
    else:
        return alldefects



from ruamel.yaml.comments import CommentedMap
from .util import NamedList
from .ends import EndList
from .tilestructures import TileStructure, getstructure
import copy


class Tile(dict):
    def __init__(self, val={}):
        dict.__init__(self, val)

        if 'type' in self.keys():
            self['structure'] = self['type']
            del(self['type'])

        if 'structure' in self.keys():
            self.structure = getstructure(self.get('structure', None),
                                          extra=self.get('extra', None))
        if self.get('extra'):
            del(self['extra'])
        
    def structure():
        doc = """The structure of the tile, set either as a string (to be looked up
with getstructure) or a TileStructure.  Get always returns the TileStructure."""
        
        def fget(self):
            return self._structure
    
        def fset(self, value):
            if value is None:
                del self['structure']
            elif isinstance(value, TileStructure):
                self._structure = value
                self['structure'] = value.name
            else:
                self._structure = getstructure(value)
                self['structure'] = self._structure.name
    
        def fdel(self):
            self._structure = None
            del self['structure']
        
        return locals()
    structure = property(**structure())

    def copy(self):
        return copy.deepcopy(self)
    
    @property
    def rotations(self):
        rl = self.structure.rotations
        tl = TileList()
        for ri, (structure, endorder) in enumerate(rl):
            t = copy.copy(self)
            if 'input' in t.keys():
                del(t['input'])
            t['ends'] = [t.ends[i] for i in endorder]
            t['rotation'] = ri
            t.structure = structure()
            t.structure._endtypes = [self.structure._endtypes[i]
                                     for i in endorder]
            tl.append(t)
        return tl


    def named_rotations(self):
        rl = self.structure.rotations
        tl = TileList()
        for ri, (structure, endorder) in enumerate(rl):
            t = copy.copy(self)
            if 'input' in t.keys():
                del(t['input'])
            t['ends'] = [t.ends[i] for i in endorder]
            t['rotation'] = ri
            t['name'] += '_rot_{}'.format(ri)
            t.structure = structure()
            t.structure._endtypes = [self.structure._endtypes[i]
                                     for i in endorder]
            tl.append(t)
        return tl
    
    
    def strands():
        doc = """Doc string"""
        
        def fget(self):
            return self.get('fullseqs', None)
    
        def fset(self, value):
            # Make sure there are the right number of ends.
            # FIXME: do a better job checking, raise a better error
            #if self.structure:
            #    assert len(self.structure._endlocs) == len(value)
            self['fullseqs'] = value
    
        def fdel(self):
            del self['fullseqs']
        return locals()
    strands = property(**strands())

    def check_strands(self):
        self.structure.check_strands(self)

    def sequence_diagram(self):
        return self.structure.sequence_diagram(self)

    def abstract_diagram(self, tileset=None):
        return self.structure.abstract_diagram(self, tileset)
        
    def name():
        doc = """Doc string"""
        def fget(self):
            return self['name']
    
        def fset(self, value):
            self['name'] = value
    
        return locals()
    name = property(**name())
    
    def ends():
        doc = """Doc string"""
        
        def fget(self):
            return self['ends']
    
        def fset(self, value):
            # Make sure there are the right number of ends.
            # FIXME: do a better job checking, raise a better error
            if self.structure:
                assert len(self.structure._endlocs) == len(value)
            self['ends'] = value
    
        def fdel(self):
            del self['ends']
        return locals()
    ends = property(**ends())

    @property
    def endlist(self):
        return self.structure.get_endlist(self)

    @property
    def orderableseqs(self):
        return self.structure.orderableseqs(self)
    
    def check_consistent(self):
        if self.structure:
            self.structure.check_consistent()
            if self.ends:
                assert(self.structure.numends == len(self.ends))
            if self.strands:
                self.check_strands()

    def check_sequences(self):
        self.structure.check_strands(self.strands)

    def __deepcopy__(self, memo):
        c = self.__class__()
        memo[id(self)] = c
        for k, v in self.items():
            c[k] = copy.deepcopy(v)
        if self.structure:
            c._structure = copy.deepcopy(self._structure)
        return c


class TileList(NamedList):
    def __init__(self, val=[]):
        NamedList.__init__(self, val)
        for i, tile in enumerate(self):
            self[i] = Tile(tile)

    def check_consistent(self):
        NamedList.check_consistent(self)
        for tile in self:
            tile.check_consistent()

    def endlist(self, fail_immediate=True):
        """Extract sticky ends from the list of tiles.

    Parameters
    ----------

    tilelist: either a list of tile description dicts, or a list of tile instances.
        If 

    fail_immediate: (default True) if True, immediately fail on a failure,
        with ValueError( tilename, exception ) from exception  If False, collect 
        exceptions, then raise ValueError( "message", [(tilename, exception)] , 
        output ).

        """
        endlist = EndList()
        errors = []

        for tile in self:
            try:
                endlist = endlist.merge(tile.endlist, in_place=True)
            except BaseException as e:
                if fail_immediate:
                    raise ValueError(tile.name) from e
                else:
                    errors.append((tile.name, e))

        if errors:
            raise ValueError("End list generation failed on:", errors, endlist)

        return endlist

from ruamel.yaml.representer import RoundTripRepresenter
RoundTripRepresenter.add_representer(TileList,
                                     RoundTripRepresenter.represent_list)
RoundTripRepresenter.add_representer(Tile,
                                     RoundTripRepresenter.represent_dict)



# New sensitivity code.
from .tilestructures import tile_daoe_single
from .tiles import TileList, Tile
from collections import Counter
from .util import comp


class CounterSet(Counter):
    def add(self, val, valthrowaway):
        self[val] += 1


class SetSet(set):
    def add(self, val, valthrowaway):
        set.add(self, val)
        
class PathSet(dict):
    def add(self, key, val):
        if key not in self.keys():
            self[key] = set()
        self[key].add(val)

def _fakesingle(tile):
    if not tile.structure.double:
        return TileList([tile])
    ret = TileList()
    c = ['', '/']
    cn = ['_db_a','_db_b']
    for qq, es in enumerate(tile.structure.singleends):
        ft = Tile()
        ft.structure = tile_daoe_single()
        ft.structure._endlocs = [None, None, None, None]
        ft['ends'] = []
        ft['name'] = tile.name+cn[qq]
        if 'input' in tile.keys():
            ft['input'] = []
        ft.structure._endtypes = []
        for i in es:
            if i is not None:
                ft['ends'].append(tile['ends'][i])
                ft.structure._endtypes.append(tile.structure._endtypes[i])
                if 'input' in tile.keys():
                    ft['input'].append(tile['input'][i])
            else:
                ft['ends'].append(tile.name + '_db' + c[qq])
                ft.structure._endtypes.append('fakedouble')
                if 'input' in tile.keys():
                    ft['input'].append(0)
        ret.append(ft)
    return ret


def _fakesingles(tiles):
    return sum((_fakesingle(x) for x in tiles), TileList())


_rev = [2, 3, 0, 1]


def sensitivity_classes(tileset, count=False, _maxorder=2):

    singles = _fakesingles(tileset.tiles)
    rotatedsingles = singles + _fakesingles(
        sum([x.rotations for x in tileset.tiles], TileList()))

    if count == 'paths':
        sclass = PathSet
    elif count:
        sclass = CounterSet
    else:
        sclass = SetSet

    spairs = {
        '1GO': sclass(),
        '1NGO': sclass(),
        '2NGO': sclass(),
        '2GO': sclass()
    }

    if _maxorder >= 3:
        spairs['22NGO'] = sclass()
        spairs['22GO'] = sclass()
        maxc = '22GO'
    else:
        maxc = '2GO'

    for t1 in singles:
        for i in range(0, 4):
            if t1.structure._endtypes[i] in {'fakedouble', 'hairpin'}:
                continue
            for t2 in rotatedsingles:
                if t2['ends'][i] != t1['ends'][i]:
                    continue
                for j in range(0, 4):
                    if i == j:
                        continue
                    if frozenset((t2['ends'][j],
                                  t1['ends'][j])) in spairs[maxc]:
                        continue
                        # don't go through the rest of the process
                        # if the pair is already in the "worst"
                        # class.
                    if t1['ends'][j] == t2['ends'][j]:
                        continue
                    if t1.structure._endtypes[j] != t2.structure._endtypes[j]:
                        continue
                    if t1.structure._endtypes[j] in {'fakedouble', 'hairpin'}:
                        continue
                    spairs['1NGO'].add(
                        frozenset((t2['ends'][j], t1['ends'][j])),
                        (t1.name, t2.name))
                    spairs['1NGO'].add(
                        frozenset((comp(t2['ends'][j]), comp(t1['ends'][j]))),
                        (t1.name, t2.name))
                    go1 = False
                    if t1['input'][i] and t1['input'][j]:
                        go1 = True
                        spairs['1GO'].add(
                            frozenset((t2['ends'][j], t1['ends'][j])),
                            (t1.name, t2.name))
                        spairs['1GO'].add(
                            frozenset((comp(t2['ends'][j]), comp(
                                t1['ends'][j]))),
                            (t1.name, t2.name))
                    for k in (set(range(0, 4)) - {i, j}):
                        if (t1.structure._endtypes[k] == 'hairpin') or (
                                t2.structure._endtypes[k] == 'hairpin'):
                            continue
                        ec1 = comp(t1['ends'][k])
                        ec2 = comp(t2['ends'][k])
                        kc = _rev[k]
                        for t12 in singles:
                            if t12['ends'][kc] != ec1:
                                continue
                            for t22 in rotatedsingles:
                                if t22['ends'][kc] != ec2:
                                    continue
                                for m in range(0, 4):
                                    if m == kc:
                                        continue
                                    if t22['ends'][m] != t12['ends'][m]:
                                        continue
                                    if t12.structure._endtypes[m] in {
                                            'fakedouble', 'hairpin'
                                    }:
                                        continue
                                    spairs['2NGO'].add(
                                        frozenset((t2['ends'][j], t1['ends'][j]
                                                   )),
                                        (t1.name, t2.name, t12.name, t22.name))
                                    spairs['2NGO'].add(
                                        frozenset((comp(t2['ends'][j]), comp(
                                            t1['ends'][j]))),
                                        (t1.name, t2.name, t12.name, t22.name))
                                    if t12['input'][m] and t12['input'][kc] and go1:
                                        spairs['2GO'].add(
                                            frozenset((t2['ends'][j], t1[
                                                'ends'][j])),
                                            (t1.name, t2.name, t12.name, t22.name))
                                        spairs['2GO'].add(
                                            frozenset((comp(t2['ends'][j]),
                                                       comp(t1['ends'][j]))),
                                            (t1.name, t2.name, t12.name, t22.name))
                                    if _maxorder > 2:
                                        for k2 in (set(range(0, 4)) - {
                                                i, j, k
                                        }):  # should just be one number!
                                            # hairpin? then continue.
                                            if (t1.structure._endtypes[k2] ==
                                                    'hairpin'
                                                ) or (
                                                    t2.structure._endtypes[k2]
                                                    == 'hairpin'):
                                                continue
                                            # fakedouble on either: then assume we're doomed: (FIXME)
                                            if (t1.structure._endtypes[k2] ==
                                                    'fakedouble'
                                                ) or (
                                                    t2.structure._endtypes[k2]
                                                    == 'fakedouble'):
                                                spairs['22NGO'].add(
                                                    frozenset((
                                                        t2['ends'][
                                                            j],
                                                        t1['ends'][
                                                            j])),
                                                    (t1.name, t2.name, t12.name, t22.name))
                                                spairs['22NGO'].add(
                                                    frozenset((comp(
                                                        t2['ends'][
                                                            j]
                                                    ), comp(
                                                        t1['ends'][
                                                            j]))),
                                                    (t1.name, t2.name, t12.name, t22.name))
                                                spairs['22GO'].add(
                                                    frozenset((
                                                        t2['ends'][
                                                            j],
                                                        t1['ends'][
                                                            j])),
                                                    (t1.name, t2.name, t12.name, t22.name))
                                                spairs['22GO'].add(
                                                    frozenset((comp(
                                                        t2['ends'][
                                                            j]
                                                    ), comp(
                                                        t1['ends'][
                                                            j]))),
                                                    (t1.name, t2.name, t12.name, t22.name))
                                                continue

                                            ec1_3 = comp(t1['ends'][k2])
                                            ec2_3 = comp(t2['ends'][k2])
                                            kc2 = _rev[k2]
                                            for t13 in singles:
                                                if t13['ends'][kc2] != ec1_3:
                                                    continue
                                                for t23 in rotatedsingles:
                                                    if t23['ends'][kc2] != ec2_3:
                                                        continue
                                                    for m2 in range(0, 4):
                                                        if m2 == kc2:
                                                            continue
                                                        if t23['ends'][m2] != t13['ends'][m2]:
                                                            continue
                                                        if t13.structure._endtypes[
                                                                m2] in {
                                                                    'fakedouble',
                                                                    'hairpin',
                                                                }:
                                                            continue
                                                        spairs['22NGO'].add(
                                                            frozenset((t2[
                                                                'ends'][j], t1[
                                                                    'ends'][j]
                                                                       )),
                                                            (t1.name, t2.name, t12.name, t22.name, t13.name, t23.name))
                                                        spairs['22NGO'].add(
                                                            frozenset((comp(
                                                                t2['ends']
                                                                [j]), comp(
                                                                    t1['ends'][
                                                                        j]))),
                                                            (t1.name, t2.name, t12.name, t22.name, t13.name, t23.name))
                                                        if t12['input'][m] and t12['input'][kc] and t13['input'][m2] and t13['input'][kc2] and go1:
                                                            spairs['22GO'].add(
                                                                frozenset((
                                                                    t2['ends'][
                                                                        j],
                                                                    t1['ends'][
                                                                        j])),
                                                                (t1.name, t2.name, t12.name, t22.name, t13.name, t23.name))
                                                            spairs['22GO'].add(
                                                                frozenset((comp(
                                                                    t2['ends'][
                                                                        j]
                                                                ), comp(
                                                                    t1['ends'][
                                                                        j]))),
                                                                (t1.name, t2.name, t12.name, t22.name, t13.name, t23.name))

    return spairs



#!/usr/bin/env python
from setuptools import setup

setup(
    name="alhambra",
    version="1.0.0-alpha.1",
    packages=['alhambra'],
    install_requires=[
        'numpy', 'stickydesign >= 0.8.1', 'svgwrite', 'lxml', 'shutilwhich',
        'peppercompiler >= 0.1.2', 'ruamel.yaml', 'cssutils'
    ],
    include_package_data=True,
    entry_points={'console_scripts': ['alhambra = alhambra.scripts:alhambra']},
    author="Constantine Evans",
    author_email="cgevans@evans.foundation",
    description="DX Tile Set Designer", )



import ruamel.yaml as yaml
import argparse
import logging
import sys
import os.path
from .tilesets import TileSet
import logging
import click

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

POOL = None

@click.group(context_settings=CONTEXT_SETTINGS, chain=True)
@click.argument('filename', type=click.Path(exists=True))
@click.option('--verbose/--quiet', '-v/-q', default=False)
@click.option(
    '--output',
    '-o',
    type=click.Path(exists=False),
    default=None,
    help='output file name (defaults to stdout)')
def alhambra(filename, verbose, output):
    pass


@alhambra.resultcallback()
def process_pipeline(processors, filename, verbose, output):
    if verbose:
        logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    ts = TileSet.from_file(filename)
    for processor in processors:
        ts = processor(ts)
    if output is None:
        ts.to_file(sys.stdout)
    else:
        ts.to_file(output)


@alhambra.command(short_help='generate sticky end sequences')
@click.option('--method', '-m', type=click.Choice(['default', 'multimodel']), help='stickydesign method to use (multimodel or default)', default='default')
@click.option('--trials', '-n', default=100, help='number of trials (defaults to 100)')
def create_end_sequences(method, trials):
    """generates sticky end sequences for the system."""
    def processor(ts):
        ts_new, _ = ts.create_end_sequences(method=method, trials=trials)
        return ts_new
    return processor


@alhambra.command(short_help='attempt to reduce the # of sticky ends')
@click.option('--latticedefects/--no-latticedefects', '-l/-L', default=True)
@click.option('--sclass', '-s', multiple=True, default=['2GO'])
@click.option('--trials', '-n', default=1)
def reduce_ends(latticedefects, sclass, trials):
    def processor(ts):
        ts_new = ts.reduce_ends(checkld=latticedefects, _classes=tuple(sclass))
        return ts_new
    return processor

@alhambra.command(short_help='attempt to reduce the # of tiles')
@click.option('--latticedefects/--no-latticedefects', '-l/-L', default=True)
@click.option('--sclass', '-s', multiple=True, default=['2GO'])
@click.option('--rotate/--no-rotate', '-r/-R', default=True)
def reduce_tiles(latticedefects, sclass, rotate):
    def processor(ts):
        ts_new = ts.reduce_tiles(checkld=latticedefects, _classes=tuple(sclass),
                                 rotation=rotate)
        return ts_new
    return processor

@alhambra.command(short_help='optimize sticky end assignment')
def reorder_ends():
    pass

@alhambra.command(short_help='create strand sequences')
def create_strand_sequences():
    pass

@alhambra.command(short_help='run xgrow')
def run_xgrow():
    pass
